overlay stdlib https://github.com/SkySkimmer/stdlib no-opt-schemes 21241

overlay bbv https://github.com/SkySkimmer/bbv no-opt-schemes 21241

overlay kami https://github.com/SkySkimmer/kami no-opt-schemes 21241
# Make PRs against https://github.com/mit-plv/kami base branch rv32i

overlay bedrock2 https://github.com/SkySkimmer/bedrock2 no-opt-schemes 21241
# Make PRs against https://github.com/mit-plv/bedrock2 base branch master

overlay coq_dpdgraph https://github.com/SkySkimmer/coq-dpdgraph no-opt-schemes 21241
