overlay metarocq https://github.com/SkySkimmer/metarocq mip-mind-record 21069

overlay lean_importer https://github.com/SkySkimmer/rocq-lean-import mip-mind-record 21069

overlay elpi https://github.com/SkySkimmer/coq-elpi mip-mind-record 21069

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp mip-mind-record 21069
