overlay lean_importer https://github.com/SkySkimmer/rocq-lean-import timeout-exns 21149

overlay waterproof https://github.com/SkySkimmer/coq-waterproof timeout-exns 21149

overlay tactician https://github.com/SkySkimmer/coq-tactician timeout-exns 21149
