overlay coq_lsp https://github.com/ppedrot/coq-lsp proof-solve-explicit-env 21110

overlay equations https://github.com/ppedrot/Coq-Equations proof-solve-explicit-env 21110

overlay mtac2 https://github.com/ppedrot/Mtac2 proof-solve-explicit-env 21110

overlay rewriter https://github.com/ppedrot/rewriter proof-solve-explicit-env 21110

overlay tactician https://github.com/ppedrot/coq-tactician proof-solve-explicit-env 21110
