overlay elpi https://github.com/ppedrot/coq-elpi reduce-pcontext-api 21394

overlay equations https://github.com/ppedrot/Coq-Equations reduce-pcontext-api 21394

overlay mtac2 https://github.com/ppedrot/Mtac2 reduce-pcontext-api 21394

overlay paramcoq https://github.com/ppedrot/paramcoq reduce-pcontext-api 21394

overlay rewriter https://github.com/ppedrot/rewriter reduce-pcontext-api 21394

overlay metarocq https://github.com/ppedrot/metarocq reduce-pcontext-api 21394
