#[warning="-level-0-notation-not-closed"]
Notation "## x" := (S x) (at level 0).
Notation "## ##" := 0 (at level 0).
Check ## [].

#[warning="-level-0-notation-not-closed"]
Notation "## x" := (S x) (at level 0).
Notation "## ## ##" := 0 (at level 0).
Check ## [].
