overlay equations https://github.com/jrosain/Rocq-Equations elimination-constraints 21195

overlay coq_dpdgraph https://github.com/jrosain/rocq-dpdgraph elimination-constraints 21195

overlay lean_importer https://github.com/jrosain/rocq-lean-import elimination-constraints 21195

overlay mtac2 https://github.com/jrosain/Mtac2 elimination-constraints 21195

overlay paramcoq https://github.com/jrosain/paramcoq elimination-constraints 21195

overlay rewriter https://github.com/jrosain/rewriter elimination-constraints 21195

overlay tactician https://github.com/jrosain/rocq-tactician elimination-constraints 21195

overlay waterproof https://github.com/jrosain/rocq-waterproof elimination-constraints 21195

overlay coq_lsp https://github.com/jrosain/rocq-lsp elimination-constraints 21195

overlay metarocq https://github.com/jrosain/metarocq elimination-constraints 21195

overlay elpi https://github.com/TDiazT/coq-elpi elimination-constraints 21195

overlay itauto https://gitlab.inria.fr/tdiaztro/itauto elimination-constraints 21195
