overlay coq_lsp https://github.com/ppedrot/coq-lsp genredexpr-open-api 21287

overlay elpi https://github.com/ppedrot/coq-elpi genredexpr-open-api 21287

overlay tactician https://github.com/ppedrot/coq-tactician genredexpr-open-api 21287
