overlay equations https://github.com/mattam82/Coq-Equations sort-poly-flags 21419
overlay rewriter https://github.com/mattam82/rewriter sort-poly-flags 21419
overlay smtcoq https://github.com/mattam82/smtcoq sort-poly-flags 21419
overlay mtac2 https://github.com/mattam82/Mtac2 sort-poly-flags 21419
overlay elpi https://github.com/mattam82/coq-elpi sort-poly-flags 21419
overlay simple_io https://github.com/mattam82/coq-simple-io sort-poly-flags 21419
overlay paramcoq https://github.com/mattam82/paramcoq sort-poly-flags 21419
overlay waterproof https://github.com/mattam82/coq-waterproof sort-poly-flags 21419
overlay itauto https://gitlab.inria.fr/sozeau/itauto sort-poly-flags 21419
overlay metarocq https://github.com/mattam82/metacoq sort-poly-flags 21419
overlay quickchick https://github.com/mattam82/quickchick sort-poly-flags 21419
