overlay coq_lsp https://github.com/SkySkimmer/coq-lsp nolab 21174

overlay tactician https://github.com/SkySkimmer/coq-tactician nolab 21174

overlay waterproof https://github.com/SkySkimmer/coq-waterproof nolab 21174
