overlay coq_dpdgraph https://github.com/ppedrot/coq-dpdgraph track-elimination-source-constraints 21486

overlay elpi https://github.com/ppedrot/coq-elpi track-elimination-source-constraints 21486

overlay equations https://github.com/ppedrot/Coq-Equations track-elimination-source-constraints 21486

overlay lean_importer https://github.com/ppedrot/rocq-lean-import track-elimination-source-constraints 21486

overlay metarocq https://github.com/ppedrot/metarocq track-elimination-source-constraints 21486

overlay mtac2 https://github.com/ppedrot/Mtac2 track-elimination-source-constraints 21486

overlay paramcoq https://github.com/ppedrot/paramcoq track-elimination-source-constraints 21486
