overlay simple_io https://github.com/SkySkimmer/coq-simple-io print-flags 21281

overlay autosubst_ocaml https://github.com/SkySkimmer/autosubst-ocaml print-flags 21281

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp print-flags 21281

overlay coqhammer https://github.com/SkySkimmer/coqhammer print-flags 21281

overlay elpi https://github.com/SkySkimmer/coq-elpi print-flags 21281

overlay equations https://github.com/SkySkimmer/Coq-Equations print-flags 21281

overlay smtcoq https://github.com/SkySkimmer/smtcoq print-flags 21281

overlay tactician https://github.com/SkySkimmer/coq-tactician print-flags 21281

overlay vsrocq https://github.com/SkySkimmer/vsrocq print-flags 21281

overlay waterproof https://github.com/SkySkimmer/coq-waterproof print-flags 21281

overlay quickchick https://github.com/SkySkimmer/QuickChick print-flags 21281
