overlay elpi https://github.com/SkySkimmer/coq-elpi stop-anomalies 21254

overlay equations https://github.com/SkySkimmer/Coq-Equations stop-anomalies 21254

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp stop-anomalies 21254
