overlay equations https://github.com/ppedrot/Coq-Equations chase-pconstraint-context-set-II 21426

overlay lean_importer https://github.com/ppedrot/rocq-lean-import chase-pconstraint-context-set-II 21426

overlay metarocq https://github.com/ppedrot/metarocq chase-pconstraint-context-set-II 21426
