overlay equations https://github.com/ppedrot/Coq-Equations reintroduce-univ-contextset 21391

overlay lean_importer https://github.com/ppedrot/rocq-lean-import reintroduce-univ-contextset 21391

overlay mtac2 https://github.com/ppedrot/Mtac2 reintroduce-univ-contextset 21391

overlay waterproof https://github.com/ppedrot/coq-waterproof reintroduce-univ-contextset 21391

overlay elpi https://github.com/ppedrot/coq-elpi reintroduce-univ-contextset 21391

overlay rewriter https://github.com/ppedrot/rewriter reintroduce-univ-contextset 21391

overlay metarocq https://github.com/ppedrot/metarocq reintroduce-univ-contextset 21391
