overlay elpi https://github.com/SkySkimmer/coq-elpi univ-decl 21395

overlay equations https://github.com/SkySkimmer/Coq-Equations univ-decl 21395

overlay rewriter https://github.com/SkySkimmer/rewriter univ-decl 21395
