Set Warnings "-closed-notation-not-level-0".

Reserved Notation "#0 #1" (at level 30).
Reserved Notation "#0 #1 #2" (at level 40).  (* no warning, parser takes successive terminals at once *)

Reserved Notation "#20 #21 x #3 y" (x at level 30, at level 50).
Reserved Notation "#20 #21 x #34" (x at level 40, at level 50).
